@@ -1,7 +1,7 @@
 #' Zap special missings to regular R missings
 #'
 #' This is useful if you want to convert tagged missing values from SAS or
-#' Stata, or user-defined missings from SPSS, to regular R \code{NA}.
+#' Stata, or user-defined missings from SPSS, to regular R `NA`.
 #'
 #' @param x A vector or data frame
 #' @export
@@ -22,7 +22,7 @@
 #' zap_missing(x2)
 #'
 #' # You can also apply to data frames
-#' df <- tibble::data_frame(x1, x2, y = 4:1)
+#' df <- tibble::tibble(x1, x2, y = 4:1)
 #' df
 #' zap_missing(df)
 zap_missing <- function(x) {
@@ -35,7 +35,7 @@ zap_missing.default <- function(x) {
 }
 
 #' @export
-zap_missing.labelled <- function(x) {
+zap_missing.haven_labelled <- function(x) {
   x[is.na(x)] <- NA
 
   labels <- attr(x, "labels")
@@ -46,12 +46,12 @@ zap_missing.labelled <- function(x) {
 }
 
 #' @export
-zap_missing.labelled_spss <- function(x) {
-  is.na(x) <- is.na(x)
+zap_missing.haven_labelled_spss <- function(x) {
+  x[is.na(x)] <- NA
 
   attr(x, "na_values") <- NULL
   attr(x, "na_range") <- NULL
-  class(x) <- "labelled"
+  class(x) <- "haven_labelled"
 
   x
 }
