@@ -1,29 +1,33 @@
 
 #include <stdio.h>
 #include <stdlib.h>
-#include <string.h>
 #include <sys/types.h>
 #include <stdint.h>
 #include <math.h>
 #include <float.h>
 #include <time.h>
+#include <limits.h>
+#include <ctype.h>
 
 #include "../readstat.h"
 #include "../readstat_bits.h"
 #include "../readstat_iconv.h"
 #include "../readstat_convert.h"
 #include "../readstat_malloc.h"
+#include "../CKHashTable.h"
 
 #include "readstat_sav.h"
 #include "readstat_sav_compress.h"
 #include "readstat_sav_parse.h"
 #include "readstat_sav_parse_timestamp.h"
+#include "readstat_sav_parse_mr_name.h"
 
 #if HAVE_ZLIB
 #include "readstat_zsav_read.h"
 #endif
 
 #define DATA_BUFFER_SIZE    65536
+#define VERY_LONG_STRING_MAX_LENGTH INT_MAX
 
 /* Others defined in table below */
 
@@ -51,9 +55,9 @@ static readstat_charset_entry_t _charset_table[] = {
     { .code = 866,   .name = "CP866" },
     { .code = 869,   .name = "CP869" },
     { .code = 874,   .name = "CP874" },
-    { .code = 932,   .name = "SHIFT-JIS" },
-    { .code = 936,   .name = "ISO-IR-58" },
-    { .code = 949,   .name = "ISO-IR-149" },
+    { .code = 932,   .name = "CP932" },
+    { .code = 936,   .name = "CP936" },
+    { .code = 949,   .name = "CP949" },
     { .code = 950,   .name = "BIG-5" },
     { .code = 1200,  .name = "UTF-16LE" },
     { .code = 1201,  .name = "UTF-16BE" },
@@ -141,7 +145,34 @@ static readstat_error_t sav_parse_machine_floating_point_record(const void *data
 static readstat_error_t sav_store_variable_display_parameter_record(const void *data, size_t size, size_t count, sav_ctx_t *ctx);
 static readstat_error_t sav_parse_variable_display_parameter_record(sav_ctx_t *ctx);
 static readstat_error_t sav_parse_machine_integer_info_record(const void *data, size_t data_len, sav_ctx_t *ctx);
-static readstat_error_t sav_parse_long_value_labels_record(const void *data, size_t size, size_t count, sav_ctx_t *ctx);
+static readstat_error_t sav_parse_long_string_value_labels_record(const void *data, size_t size, size_t count, sav_ctx_t *ctx);
+static readstat_error_t sav_parse_long_string_missing_values_record(const void *data, size_t size, size_t count, sav_ctx_t *ctx);
+static readstat_error_t sav_read_multiple_response_sets(size_t data_len, sav_ctx_t *ctx);
+
+static readstat_error_t sav_read_multiple_response_sets(size_t data_len, sav_ctx_t *ctx) {
+    readstat_error_t retval = READSTAT_OK;
+
+    char *mr_string = readstat_malloc(data_len + 1);
+    if (mr_string == NULL) {
+        retval = READSTAT_ERROR_MALLOC;
+        goto cleanup;
+    }
+    mr_string[data_len] = '\0';
+    if (ctx->io->read(mr_string, data_len, ctx->io->io_ctx) < data_len) {
+        retval = READSTAT_ERROR_PARSE;
+        goto cleanup;
+    }
+    if (mr_string[0] != '$') {
+        retval = READSTAT_ERROR_BAD_MR_STRING;
+        goto cleanup;
+    }
+
+    retval = parse_mr_string(mr_string, &ctx->mr_sets, &ctx->multiple_response_sets_length);
+
+cleanup:
+    free(mr_string);
+    return retval;
+}
 
 static void sav_tag_missing_double(readstat_value_t *value, sav_ctx_t *ctx) {
     double fp_value = value->v.double_value;
@@ -334,7 +365,7 @@ static readstat_error_t sav_read_variable_record(sav_ctx_t *ctx) {
         }
         ctx->var_offset++;
         ctx->varinfo[ctx->var_index-1]->width++;
-        return 0;
+        return retval;
     }
 
     if ((info = readstat_calloc(1, sizeof(spss_varinfo_t))) == NULL) {
@@ -348,12 +379,12 @@ static readstat_error_t sav_read_variable_record(sav_ctx_t *ctx) {
     info->labels_index = -1;
 
     retval = readstat_convert(info->name, sizeof(info->name),
-            variable.name, sizeof(variable.name), ctx->converter);
+            variable.name, sizeof(variable.name), NULL);
     if (retval != READSTAT_OK)
         goto cleanup;
 
     retval = readstat_convert(info->longname, sizeof(info->longname), 
-            variable.name, sizeof(variable.name), ctx->converter);
+            variable.name, sizeof(variable.name), NULL);
     if (retval != READSTAT_OK)
         goto cleanup;
 
@@ -677,6 +708,11 @@ static readstat_error_t sav_read_dictionary_termination_record(sav_ctx_t *ctx) {
 }
 
 static readstat_error_t sav_process_row(unsigned char *buffer, size_t buffer_len, sav_ctx_t *ctx) {
+    if (ctx->row_offset) {
+        ctx->row_offset--;
+        return READSTAT_OK;
+    }
+
     readstat_error_t retval = READSTAT_OK;
     double fp_value;
     int offset = 0;
@@ -684,6 +720,7 @@ static readstat_error_t sav_process_row(unsigned char *buffer, size_t buffer_len
     size_t raw_str_used = 0;
     int segment_offset = 0;
     int var_index = 0, col = 0;
+    int raw_str_is_utf8 = ctx->input_encoding && !strcmp(ctx->input_encoding, "UTF-8");
 
     while (data_offset < buffer_len && col < ctx->var_index && var_index < ctx->var_index) {
         spss_varinfo_t *col_info = ctx->varinfo[col];
@@ -694,16 +731,25 @@ static readstat_error_t sav_process_row(unsigned char *buffer, size_t buffer_len
             goto done;
         }
         if (var_info->type == READSTAT_TYPE_STRING) {
-            if (raw_str_used + 8 <= ctx->raw_string_len) {
-                memcpy(ctx->raw_string + raw_str_used, &buffer[data_offset], 8);
-                raw_str_used += 8;
+            // If we're in the last column of a segment, only read 7 bytes
+            // (Segments contain 255 bytes but have room for 256)
+            size_t read_len = 8 - (offset == 31);
+            if (raw_str_used + read_len <= ctx->raw_string_len) {
+                if (raw_str_is_utf8) {
+                    /* Skip null bytes, see https://github.com/tidyverse/haven/issues/560  */
+                    char c;
+                    for (int i=0; i<read_len; i++)
+                        if ((c = buffer[data_offset+i]))
+                            ctx->raw_string[raw_str_used++] = c;
+                } else {
+                    memcpy(ctx->raw_string + raw_str_used, &buffer[data_offset], read_len);
+                    raw_str_used += read_len;
+                }
             }
             if (++offset == col_info->width) {
-                if (++segment_offset < var_info->n_segments) {
-                    raw_str_used--;
-                }
                 offset = 0;
                 col++;
+                segment_offset++;
             }
             if (segment_offset == var_info->n_segments) {
                 if (!ctx->variables[var_info->index]->skip) {
@@ -784,7 +830,7 @@ static readstat_error_t sav_read_data(sav_ctx_t *ctx) {
     if (retval != READSTAT_OK)
         goto done;
 
-    if (ctx->record_count != -1 && ctx->current_row != ctx->row_limit) {
+    if (ctx->record_count >= 0 && ctx->current_row != ctx->row_limit) {
         retval = READSTAT_ERROR_ROW_COUNT_MISMATCH;
     }
 
@@ -802,6 +848,14 @@ static readstat_error_t sav_read_uncompressed_data(sav_ctx_t *ctx,
 
     buffer = readstat_malloc(buffer_len);
 
+    if (ctx->row_offset) {
+        if (io->seek(buffer_len * ctx->row_offset, READSTAT_SEEK_CUR, io->io_ctx) == -1) {
+            retval = READSTAT_ERROR_SEEK;
+            goto done;
+        }
+        ctx->row_offset = 0;
+    }
+
     while (ctx->row_limit == -1 || ctx->current_row < ctx->row_limit) {
         retval = sav_update_progress(ctx);
         if (retval != READSTAT_OK)
@@ -844,6 +898,10 @@ static readstat_error_t sav_read_compressed_data(sav_ctx_t *ctx,
     }
 
     while (1) {
+        retval = sav_update_progress(ctx);
+        if (retval != READSTAT_OK)
+            goto done;
+
         buffer_used = io->read(buffer, sizeof(buffer), io->io_ctx);
         if (buffer_used == -1 || buffer_used == 0 || (buffer_used % 8) != 0)
             goto done;
@@ -871,7 +929,9 @@ static readstat_error_t sav_read_compressed_data(sav_ctx_t *ctx,
                 uncompressed_offset = 0;
             }
 
-            if (state.status == SAV_ROW_STREAM_FINISHED_ALL || ctx->current_row == ctx->row_limit)
+            if (state.status == SAV_ROW_STREAM_FINISHED_ALL)
+                goto done;
+            if (ctx->row_limit > 0 && ctx->current_row == ctx->row_limit)
                 goto done;
         }
     }
@@ -914,7 +974,13 @@ static readstat_error_t sav_parse_machine_integer_info_record(const void *data,
         }
         ctx->input_encoding = src_charset;
     }
-    if (src_charset && dst_charset && strcmp(src_charset, dst_charset) != 0) {
+    if (src_charset && dst_charset) {
+        // You might be tempted to skip the charset conversion when src_charset
+        // and dst_charset are the same. However, some versions of SPSS insert
+        // illegally truncated strings (e.g. the last character is three bytes
+        // but the field only has room for two bytes). So to prevent the client
+        // from receiving an invalid byte sequence, we ram everything through
+        // our iconv machinery.
         iconv_t converter = iconv_open(dst_charset, src_charset);
         if (converter == (iconv_t)-1) {
             return READSTAT_ERROR_UNSUPPORTED_CHARSET;
@@ -967,13 +1033,14 @@ static readstat_error_t sav_parse_variable_display_parameter_record(sav_ctx_t *c
 
     int i;
     long count = ctx->variable_display_values_count;
-    if (count != 2 * ctx->var_count && count != 3 * ctx->var_count) {
+    if (count != 2 * ctx->var_index && count != 3 * ctx->var_index) {
         return READSTAT_ERROR_PARSE;
     }
-    int has_display_width = ctx->var_count > 0 && (count / ctx->var_count == 3);
+    int has_display_width = ctx->var_index > 0 && (count / ctx->var_index == 3);
     int offset = 0;
     for (i=0; i<ctx->var_index;) {
         spss_varinfo_t *info = ctx->varinfo[i];
+        offset = (2 + has_display_width)*i;
         info->measure = spss_measure_to_readstat_measure(ctx->variable_display_values[offset++]);
         if (has_display_width) {
             info->display_width = ctx->variable_display_values[offset++];
@@ -985,19 +1052,53 @@ static readstat_error_t sav_parse_variable_display_parameter_record(sav_ctx_t *c
     return READSTAT_OK;
 }
 
-static readstat_error_t sav_parse_long_value_labels_record(const void *data, size_t size, size_t count, sav_ctx_t *ctx) {
+static readstat_error_t sav_read_pascal_string(char *buf, size_t buf_len,
+        const char **inout_data_ptr, size_t data_ptr_len, sav_ctx_t *ctx) {
+    const char *data_ptr = *inout_data_ptr;
+    const char *data_end = data_ptr + data_ptr_len;
+    readstat_error_t retval = READSTAT_OK;
+    uint32_t var_name_len = 0;
+
+    if (data_ptr + sizeof(uint32_t) > data_end) {
+        retval = READSTAT_ERROR_PARSE;
+        goto cleanup;
+    }
+
+    memcpy(&var_name_len, data_ptr, sizeof(uint32_t));
+    if (ctx->bswap)
+        var_name_len = byteswap4(var_name_len);
+
+    data_ptr += sizeof(uint32_t);
+
+    if (data_ptr + var_name_len > data_end) {
+        retval = READSTAT_ERROR_PARSE;
+        goto cleanup;
+    }
+
+    retval = readstat_convert(buf, buf_len, data_ptr, var_name_len, NULL);
+    if (retval != READSTAT_OK)
+        goto cleanup;
+
+    data_ptr += var_name_len;
+
+cleanup:
+    *inout_data_ptr = data_ptr;
+
+    return retval;
+}
+
+static readstat_error_t sav_parse_long_string_value_labels_record(const void *data, size_t size, size_t count, sav_ctx_t *ctx) {
     if (!ctx->handle.value_label)
         return READSTAT_OK;
     if (size != 1)
         return READSTAT_ERROR_PARSE;
 
     readstat_error_t retval = READSTAT_OK;
-    uint32_t label_name_len = 0;
     uint32_t label_count = 0;
     uint32_t i = 0;
     const char *data_ptr = data;
     const char *data_end = data_ptr + count;
-    char var_name_buf[256*4+1];
+    char var_name_buf[256+1]; // unconverted
     char label_name_buf[256];
     char *value_buffer = NULL;
     char *label_buffer = NULL;
@@ -1005,28 +1106,11 @@ static readstat_error_t sav_parse_long_value_labels_record(const void *data, siz
     while (data_ptr < data_end) {
         memset(label_name_buf, '\0', sizeof(label_name_buf));
 
-        if (data_ptr + sizeof(uint32_t) > data_end) {
-            retval = READSTAT_ERROR_PARSE;
-            goto cleanup;
-        }
-
-        memcpy(&label_name_len, data_ptr, sizeof(uint32_t));
-        if (ctx->bswap)
-            label_name_len = byteswap4(label_name_len);
-
-        data_ptr += sizeof(uint32_t);
-
-        if (data_ptr + label_name_len > data_end) {
-            retval = READSTAT_ERROR_PARSE;
-            goto cleanup;
-        }
-
-        retval = readstat_convert(var_name_buf, sizeof(var_name_buf), data_ptr, label_name_len, ctx->converter);
+        retval = sav_read_pascal_string(var_name_buf, sizeof(var_name_buf),
+                &data_ptr, data_end - data_ptr, ctx);
         if (retval != READSTAT_OK)
             goto cleanup;
 
-        data_ptr += label_name_len;
-
         for (i=0; i<ctx->var_index;) {
             spss_varinfo_t *info = ctx->varinfo[i];
             if (strcmp(var_name_buf, info->longname) == 0) {
@@ -1140,6 +1224,83 @@ cleanup:
     return retval;
 }
 
+static readstat_error_t sav_parse_long_string_missing_values_record(const void *data, size_t size, size_t count, sav_ctx_t *ctx) {
+    if (size != 1)
+        return READSTAT_ERROR_PARSE;
+
+    readstat_error_t retval = READSTAT_OK;
+    uint32_t i = 0, j = 0;
+    const char *data_ptr = data;
+    const char *data_end = data_ptr + count;
+    char var_name_buf[256+1];
+
+    while (data_ptr < data_end) {
+        retval = sav_read_pascal_string(var_name_buf, sizeof(var_name_buf),
+                &data_ptr, data_end - data_ptr, ctx);
+        if (retval != READSTAT_OK)
+            goto cleanup;
+
+        if (data_ptr == data_end) {
+            retval = READSTAT_ERROR_PARSE;
+            goto cleanup;
+        }
+
+        char n_missing_values = *data_ptr++;
+        if (n_missing_values < 1 || n_missing_values > 3) {
+            retval = READSTAT_ERROR_PARSE;
+            goto cleanup;
+        }
+
+        for (i=0; i<ctx->var_index;) {
+            spss_varinfo_t *info = ctx->varinfo[i];
+            if (strcmp(var_name_buf, info->longname) == 0) {
+                info->n_missing_values = n_missing_values;
+
+                uint32_t var_name_len = 0;
+
+                if (data_ptr + sizeof(uint32_t) > data_end) {
+                    retval = READSTAT_ERROR_PARSE;
+                    goto cleanup;
+                }
+
+                memcpy(&var_name_len, data_ptr, sizeof(uint32_t));
+                if (ctx->bswap)
+                    var_name_len = byteswap4(var_name_len);
+
+                data_ptr += sizeof(uint32_t);
+
+                for (j=0; j<n_missing_values; j++) {
+                    if (data_ptr + var_name_len > data_end) {
+                        retval = READSTAT_ERROR_PARSE;
+                        goto cleanup;
+                    }
+
+                    retval = readstat_convert(info->missing_string_values[j],
+                            sizeof(info->missing_string_values[0]),
+                            data_ptr, var_name_len, ctx->converter);
+                    if (retval != READSTAT_OK)
+                        goto cleanup;
+
+                    data_ptr += var_name_len;
+                }
+                break;
+            }
+            i += info->n_segments;
+        }
+        if (i == ctx->var_index) {
+            retval = READSTAT_ERROR_PARSE;
+            goto cleanup;
+        }
+    }
+
+    if (data_ptr != data_end) {
+        retval = READSTAT_ERROR_PARSE;
+    }
+
+cleanup:
+    return retval;
+}
+
 static readstat_error_t sav_parse_records_pass1(sav_ctx_t *ctx) {
     char data_buf[4096];
     readstat_error_t retval = READSTAT_OK;
@@ -1203,6 +1364,14 @@ static readstat_error_t sav_parse_records_pass1(sav_ctx_t *ctx) {
                     retval = sav_parse_machine_integer_info_record(data_buf, data_len, ctx);
                     if (retval != READSTAT_OK)
                         goto cleanup;
+                } else if (subtype == SAV_RECORD_SUBTYPE_MULTIPLE_RESPONSE_SETS) {
+                    if (ctx->mr_sets != NULL) {
+                        retval = READSTAT_ERROR_BAD_MR_STRING;
+                        goto cleanup;
+                    }
+                    retval = sav_read_multiple_response_sets(data_len, ctx);
+                    if (retval != READSTAT_OK)
+                        goto cleanup;
                 } else {
                     if (io->seek(data_len, READSTAT_SEEK_CUR, io->io_ctx) == -1) {
                         retval = READSTAT_ERROR_SEEK;
@@ -1314,10 +1483,16 @@ static readstat_error_t sav_parse_records_pass2(sav_ctx_t *ctx) {
                         if (retval != READSTAT_OK)
                             goto cleanup;
                         break;
-                    case SAV_RECORD_SUBTYPE_LONG_VALUE_LABELS:
-                        retval = sav_parse_long_value_labels_record(data_buf, size, count, ctx);
+                    case SAV_RECORD_SUBTYPE_LONG_STRING_VALUE_LABELS:
+                        retval = sav_parse_long_string_value_labels_record(data_buf, size, count, ctx);
                         if (retval != READSTAT_OK)
                             goto cleanup;
+                        break;
+                    case SAV_RECORD_SUBTYPE_LONG_STRING_MISSING_VALUES:
+                        retval = sav_parse_long_string_missing_values_record(data_buf, size, count, ctx);
+                        if (retval != READSTAT_OK)
+                            goto cleanup;
+                        break;
                     default: /* misc. info */
                         break;
                 }
@@ -1336,11 +1511,13 @@ cleanup:
     return retval;
 }
 
-static void sav_set_n_segments_and_var_count(sav_ctx_t *ctx) {
+static readstat_error_t sav_set_n_segments_and_var_count(sav_ctx_t *ctx) {
     int i;
     ctx->var_count = 0;
     for (i=0; i<ctx->var_index;) {
         spss_varinfo_t *info = ctx->varinfo[i];
+        if (info->string_length > VERY_LONG_STRING_MAX_LENGTH)
+            return READSTAT_ERROR_PARSE;
         if (info->string_length) {
             info->n_segments = (info->string_length + 251) / 252;
         }
@@ -1348,6 +1525,7 @@ static void sav_set_n_segments_and_var_count(sav_ctx_t *ctx) {
         i += info->n_segments;
     }
     ctx->variables = readstat_calloc(ctx->var_count, sizeof(readstat_variable_t *));
+    return READSTAT_OK;
 }
 
 static readstat_error_t sav_handle_variables(sav_ctx_t *ctx) {
@@ -1361,7 +1539,7 @@ static readstat_error_t sav_handle_variables(sav_ctx_t *ctx) {
     for (i=0; i<ctx->var_index;) {
         char label_name_buf[256];
         spss_varinfo_t *info = ctx->varinfo[i];
-        ctx->variables[info->index] = spss_init_variable_for_info(info, index_after_skipping);
+        ctx->variables[info->index] = spss_init_variable_for_info(info, index_after_skipping, ctx->converter);
 
         snprintf(label_name_buf, sizeof(label_name_buf), SAV_LABEL_NAME_PREFIX "%d", info->labels_index);
 
@@ -1464,14 +1642,23 @@ readstat_error_t readstat_parse_sav(readstat_parser_t *parser, const char *path,
     ctx->output_encoding = parser->output_encoding;
     ctx->user_ctx = user_ctx;
     ctx->file_size = file_size;
-    if (parser->row_limit > 0 && (parser->row_limit < ctx->record_count || ctx->record_count == -1)) {
+    if (parser->row_offset > 0)
+        ctx->row_offset = parser->row_offset;
+    if (ctx->record_count >= 0) {
+        int record_count_after_skipping = ctx->record_count - ctx->row_offset;
+        if (record_count_after_skipping < 0) {
+            record_count_after_skipping = 0;
+            ctx->row_offset = ctx->record_count;
+        }
+        ctx->row_limit = record_count_after_skipping;
+        if (parser->row_limit > 0 && parser->row_limit < record_count_after_skipping) 
+            ctx->row_limit = parser->row_limit;
+    } else if (parser->row_limit > 0) {
         ctx->row_limit = parser->row_limit;
-    } else {
-        ctx->row_limit = ctx->record_count;
     }
     
-    if ((retval = sav_parse_timestamp(ctx, &header)) != READSTAT_OK)
-        goto cleanup;
+    /* ignore errors */
+    sav_parse_timestamp(ctx, &header);
 
     if ((retval = sav_parse_records_pass1(ctx)) != READSTAT_OK)
         goto cleanup;
@@ -1487,7 +1674,8 @@ readstat_error_t readstat_parse_sav(readstat_parser_t *parser, const char *path,
     if ((retval = sav_parse_records_pass2(ctx)) != READSTAT_OK)
         goto cleanup;
  
-    sav_set_n_segments_and_var_count(ctx);
+    if ((retval = sav_set_n_segments_and_var_count(ctx)) != READSTAT_OK)
+        goto cleanup;
 
     if (ctx->var_count == 0) {
         retval = READSTAT_ERROR_PARSE;
@@ -1496,7 +1684,7 @@ readstat_error_t readstat_parse_sav(readstat_parser_t *parser, const char *path,
 
     if (ctx->handle.metadata) {
         readstat_metadata_t metadata = {
-            .row_count = ctx->record_count == -1 ? -1 : ctx->row_limit,
+            .row_count = ctx->record_count < 0 ? -1 : ctx->row_limit,
             .var_count = ctx->var_count,
             .file_encoding = ctx->input_encoding,
             .file_format_version = ctx->format_version,
@@ -1511,17 +1699,61 @@ readstat_error_t readstat_parse_sav(readstat_parser_t *parser, const char *path,
 
         metadata.file_label = ctx->file_label;
 
+        // Replace short MR names with long names
+        ck_hash_table_t *var_dict = ck_hash_table_init(ctx->var_index, 8);
+        for (size_t i = 0; i < ctx->var_index; i++) {
+            spss_varinfo_t *current_varinfo = ctx->varinfo[i];
+            if (current_varinfo != NULL && current_varinfo->name[0] != '\0') {
+                ck_str_hash_insert(current_varinfo->name, current_varinfo, var_dict);
+            }
+        }
+        for (size_t i = 0; i < ctx->multiple_response_sets_length; i++) {
+            mr_set_t mr = ctx->mr_sets[i];
+            for (size_t j = 0; j < mr.num_subvars; j++) {
+                char* sv_name_upper = readstat_malloc(strlen(mr.subvariables[j]) + 1);
+                if (sv_name_upper == NULL) {
+                    retval = READSTAT_ERROR_MALLOC;
+                    goto cleanup;
+                }
+                sv_name_upper[strlen(mr.subvariables[j])] = '\0';
+                for (int c = 0; mr.subvariables[j][c] != '\0'; c++) {
+                    sv_name_upper[c] = toupper((unsigned char) mr.subvariables[j][c]);
+                }
+                spss_varinfo_t *info = (spss_varinfo_t *)ck_str_hash_lookup(sv_name_upper, var_dict);
+                if (info) {
+                    free(mr.subvariables[j]);
+                    // mr.subvariables[j] = NULL;
+                    if ((mr.subvariables[j] = readstat_malloc(strlen(info->longname) + 1)) == NULL) {
+                        retval = READSTAT_ERROR_MALLOC;
+                        goto cleanup;
+                    }
+                    // mr.subvariables[j][strlen(info->longname)] = '\0';
+                    strcpy(mr.subvariables[j], info->longname);
+                    // mr.subvariables[j] = info->longname;
+                }
+                free(sv_name_upper);
+                // sv_name_upper = NULL;
+            }
+        }
+        if (var_dict)
+            ck_hash_table_free(var_dict);
+
+        metadata.multiple_response_sets_length = ctx->multiple_response_sets_length;
+        metadata.mr_sets = ctx->mr_sets;
+
         if (ctx->handle.metadata(&metadata, ctx->user_ctx) != READSTAT_HANDLER_OK) {
             retval = READSTAT_ERROR_USER_ABORT;
             goto cleanup;
         }
     }
 
-    sav_parse_variable_display_parameter_record(ctx);
+    if ((retval = sav_parse_variable_display_parameter_record(ctx)) != READSTAT_OK)
+        goto cleanup;
 
     if ((retval = sav_handle_variables(ctx)) != READSTAT_OK)
         goto cleanup;
 
+
     if ((retval = sav_handle_fweight(ctx)) != READSTAT_OK)
         goto cleanup;
 
