@@ -243,7 +243,7 @@ static int sav_n_missing_string_values(readstat_variable_t *r_variable) {
 
 static readstat_error_t sav_n_missing_values(int *out_n_missing_values, readstat_variable_t *r_variable) {
     int n_missing_values = 0;
-    if (r_variable->type == READSTAT_TYPE_DOUBLE) {
+    if (readstat_variable_get_type_class(r_variable) == READSTAT_TYPE_CLASS_NUMERIC) {
         n_missing_values = sav_n_missing_double_values(r_variable);
     } else if (readstat_variable_get_storage_width(r_variable) <= 8) {
         n_missing_values = sav_n_missing_string_values(r_variable);
@@ -344,7 +344,7 @@ cleanup:
 }
 
 static readstat_error_t sav_emit_variable_missing_values(readstat_writer_t *writer, readstat_variable_t *r_variable) {
-    if (r_variable->type == READSTAT_TYPE_DOUBLE) {
+    if (readstat_variable_get_type_class(r_variable) == READSTAT_TYPE_CLASS_NUMERIC) {
         return sav_emit_variable_missing_double_values(writer, r_variable);
     } else if (readstat_variable_get_storage_width(r_variable) <= 8) {
         return sav_emit_variable_missing_string_values(writer, r_variable);
@@ -1236,7 +1236,7 @@ static readstat_error_t sav_validate_name_chars(const char *name, int unicode) {
         if (name[j] == ' ')
             return READSTAT_ERROR_NAME_CONTAINS_ILLEGAL_CHARACTER;
 
-        if ((name[j] > 0 || !unicode) && name[j] != '@' &&
+        if (((unsigned char)name[j] < 0x80 || !unicode) && name[j] != '@' &&
                 name[j] != '.' && name[j] != '_' &&
                 name[j] != '$' && name[j] != '#' &&
                 !(name[j] >= 'a' && name[j] <= 'z') &&
@@ -1246,7 +1246,7 @@ static readstat_error_t sav_validate_name_chars(const char *name, int unicode) {
         }
     }
     char first_char = name[0];
-    if ((first_char > 0 || !unicode) && first_char != '@' &&
+    if (((unsigned char)first_char < 0x80 || !unicode) && first_char != '@' &&
             !(first_char >= 'a' && first_char <= 'z') &&
             !(first_char >= 'A' && first_char <= 'Z')) {
         return READSTAT_ERROR_NAME_BEGINS_WITH_ILLEGAL_CHARACTER;
@@ -1304,7 +1304,7 @@ static sav_varnames_t *sav_varnames_init(readstat_writer_t *writer) {
             shortname[k] = toupper(shortname[k]);
         }
         if (ck_str_hash_lookup(shortname, table)) {
-            snprintf(shortname, sizeof(varnames[0].shortname), "V%d_A", i+1);
+            snprintf(shortname, sizeof(varnames[0].shortname), "V%d_A", ((unsigned int)i+1)%100000);
         }
         ck_str_hash_insert(shortname, r_variable, table);
 
